/*
 * Copyright (c) 2020,2023 LAAS/CNRS
 * All rights reserved.
 *
 * Redistribution and use  in source  and binary  forms,  with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *   1. Redistributions of  source  code must retain the  above copyright
 *      notice and this list of conditions.
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice and  this list of  conditions in the  documentation and/or
 *      other materials provided with the distribution.
 *
 *                                    Gianluca Corsini on Mon Nov 23 2020
 */
#include "aclibdynamixel.h"

#define _XOPEN_SOURCE 600 /* linux does not have ptsname(3) otherwise */
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/uio.h>

#include <err.h>
#include <errno.h>
#include <fcntl.h>
#include <poll.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>

#include "private.h"

static uint16_t	crc16_8005(uint16_t crc, uint8_t *data, size_t size);


/* --- dynamixel_init_pty -------------------------------------------------- */

struct dynamixel_pty_s *
dynamixel_init_pty()
{
  struct dynamixel_pty_s *pty;

  pty = malloc(sizeof(*pty));
  if (!pty) return NULL;

  *pty = (struct dynamixel_pty_s){
    .fd = -1
  };

  return pty;
}


/* --- dynamixel_fini_pty -------------------------------------------------- */

/* close pseudo terminal, remove symlink */
void
dynamixel_fini_pty(struct dynamixel_pty_s *pty)
{
  if (!pty) return;

  if (pty->fd >= 0) close(pty->fd);
  unlink(pty->path);

  free(pty);
}


/* --- dynamixel_open_pty -------------------------------------------------- */

/* Open a pseudo terminal, create a symlink.
 * returns: 0 on success errno on error.
 */
int
dynamixel_open_pty(struct dynamixel_pty_s *pty, const char *path)
{
  int fd;

  /* open non-blocking */
  fd = open("/dev/ptmx", O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (fd < 0) return errno;

  /* create symlink */
  unlink(path);
  if (symlink(ptsname(fd), path)) return errno;

  /* configure */
  if (grantpt(fd) || unlockpt(fd)) return errno;

  /* user handle */
  if (pty->fd >= 0) close(pty->fd);
  pty->fd = fd;
  if (path != pty->path)
    snprintf(pty->path, sizeof(pty->path), "%s", path);
  pty->r = pty->w = pty->state = 0;

  return 0;
}


/* --- dynamixel_pty_name -------------------------------------------------- */

const char *
dynamixel_pty_name(struct dynamixel_pty_s *pty)
{
  return pty->path;
}


/* --- dynamixel_wait_msg -------------------------------------------------- */

/* returns: 1 if data present, 0 on timeout or HANGHUP, < 0 on error */
int
dynamixel_wait_msg(struct dynamixel_pty_s *pty, int timeoutms)
{
  struct pollfd pfds;
  int s;

  pfds.fd = pty->fd;
  pfds.events = POLLIN;

  do {
    s = poll(&pfds, 1, timeoutms);
  } while (s < 0 && errno == EINTR);

  if (s <= 0) return s;

  if (pfds.revents & POLLIN) return 1; /* there is data to read */
  if (pfds.revents & POLLNVAL) return -1; /* fd not open */

  if (pfds.revents & POLLHUP) { /* reopen */
    if (dynamixel_open_pty(pty, pty->path)) return -1;
    return 0;
  }

  return s;
}


/* --- ldx_recv_msg -------------------------------------------------------- */

/* returns: 0: timeout/incomplete, -1: error, 1: complete msg */

ssize_t
ldx_recv_msg(struct dynamixel_pty_s *pty)
{
  struct iovec iov[2];
  ssize_t s;
  uint8_t c;

  do {
    /* feed the ring buffer */
    iov[0].iov_base = pty->buf + pty->w;
    iov[1].iov_base = pty->buf;

    if (pty->r > pty->w) {
      iov[0].iov_len = pty->r - pty->w - 1; /* all space in iov[0] */
      iov[1].iov_len = 0;
    } else if (pty->r > 0) {
      iov[0].iov_len = sizeof(pty->buf) - pty->w;
      iov[1].iov_len = pty->r - 1;
    } else {
      iov[0].iov_len = sizeof(pty->buf) - pty->w - 1;
      iov[1].iov_len = 0;
    }

    if (iov[0].iov_len || iov[1].iov_len) {
      do {
        s = readv(pty->fd, iov, 2);
      } while(s < 0 && errno == EINTR);

      if (s <= 0 && pty->r == pty->w) return s;
      if (s > 0)
        pty->w = (pty->w + s) % sizeof(pty->buf);
    }

    while(pty->r != pty->w) {
      c = pty->buf[pty->r];
      pty->r = (pty->r + 1) % sizeof(pty->buf);

      pty->header = (pty->header << 8) | c;
      if (pty->header == 0xfffffd00) {
        if (pty->state && pty->state != 99)
          warnx("dropped partial message");
        /* found header, jump to state 1 */
        pty->state = 1; continue;
      }

      switch(pty->state) {
        case 0: pty->state = 99; break;
        case 99: break;

        case 1: /* id */
          pty->id = c;
          pty->crc = crc16_8005(0, (uint8_t []){0xff, 0xff, 0xfd, 0, c}, 5);
          pty->state++; break;

        case 2: /* len LSB */
          pty->len = c;
          pty->crc = crc16_8005(pty->crc, &c, 1);
          pty->state++; break;

        case 3: /* len MSB */
          pty->len += (c << 8);

          /* min length is 3 */
          if (pty->len < 3) {
            warnx("dropped message with bad length");
            pty->state = 0; break;
          }
          pty->len -= 3; /* len is #bytes in PARAM field */
          pty->p = 0;
          pty->crc = crc16_8005(pty->crc, &c, 1);
          pty->state++; break;

        case 4: /* cmd (instruction) */
          pty->cmd = c;
          pty->crc = crc16_8005(pty->crc, &c, 1);
          if (!pty->len)
            pty->state = 6; /* no parameters, goto to CRC case */
          else
            pty->state++;
          break;

        case 5: /* data: the parameters */
          if (pty->header != 0xfffffdfd) {
            /* dynamixel "byte stuffing": ignore the last 0xfd in data when
             * 0xfffffdfd is seen, but include it in CRC computation */
            pty->params[pty->p++] = c;
          }
          pty->crc = crc16_8005(pty->crc, &c, 1);
          pty->len--;
          if (!pty->len) { pty->state++; }
          break;

        case 6: /* crc LSB */
          pty->crc ^= c;
          pty->state++;
          break;

        case 7: /* crc MSB */
          pty->state = 0;
          pty->crc ^= (c << 8);
          return 1;
      }
    }
  } while(1);

  return 0;
}


/* --- ldx_send_status_msg ------------------------------------------------- */

/* returns normally 0, in case of errors -1 (invalid fd) */

int
ldx_send_status_msg(const struct dynamixel_pty_s *pty,
                    uint8_t id, bool alert, uint8_t error,
                    uint16_t nparams, uint8_t *params)
{
  uint8_t buf[512], *w, *r;
  uint16_t cksum, len;
  uint32_t stuff;
  ssize_t s;
  int i;

  if (!pty || pty->fd < 0) return -1;

  /* check buffer size (including byte stuffing and some overhead) */
  if (nparams > sizeof(buf)/2) return -1;

  w = buf;

  *w++ = 0xff; /* header */
  *w++ = 0xff;
  *w++ = 0xfd;
  *w++ = 0;

  *w++ = id; /* motor id */
  w += 2; /* packer length, updated later */

  *w++ = 0x55; /* status packet */
  *w++ = ((!!alert) << 7) | (error & 0x7f); /* errors */

  len = 4 + nparams;
  stuff = 0;
  for(i = 0; i < nparams; i++) { /* params */
    stuff = ((stuff << 8) & 0xffffff) | *params;
    *w++ = *params++;
    if (stuff == 0xfffffd) {
      /* "byte stuffing", add extra 0xfd and increase packet length */
      /* https://emanual.robotis.com/docs/en/dxl/protocol2/#packet-process */
      *w++ = 0xfd;
      len++;
    }
  }

  buf[5] = len & 0xff; /* packet length */
  buf[6] = len >> 8;

  cksum = crc16_8005(0, buf, w-buf); /* CRC */
  *w++ = cksum & 0xff;
  *w++ = cksum >> 8;

  /* send */
  r = buf;
  while (w > r) {
    do {
      s = write(pty->fd, r, w - r);
    } while (s < 0 && errno == EINTR);
    if (s < 0) return -1;

    r += s;
  }

  return 0;
}


/* --- crc16_8005 ---------------------------------------------------------- */

/** Compute a CRC16 cksum with polynomial x^16 + x^15 + x^2 + 1 */

static uint16_t
crc16_8005(uint16_t crc, uint8_t *data, size_t size)
{
  static const uint16_t crc_table[256] = {
    0x0000, 0x8005, 0x800F, 0x000A, 0x801B, 0x001E, 0x0014, 0x8011,
    0x8033, 0x0036, 0x003C, 0x8039, 0x0028, 0x802D, 0x8027, 0x0022,
    0x8063, 0x0066, 0x006C, 0x8069, 0x0078, 0x807D, 0x8077, 0x0072,
    0x0050, 0x8055, 0x805F, 0x005A, 0x804B, 0x004E, 0x0044, 0x8041,
    0x80C3, 0x00C6, 0x00CC, 0x80C9, 0x00D8, 0x80DD, 0x80D7, 0x00D2,
    0x00F0, 0x80F5, 0x80FF, 0x00FA, 0x80EB, 0x00EE, 0x00E4, 0x80E1,
    0x00A0, 0x80A5, 0x80AF, 0x00AA, 0x80BB, 0x00BE, 0x00B4, 0x80B1,
    0x8093, 0x0096, 0x009C, 0x8099, 0x0088, 0x808D, 0x8087, 0x0082,
    0x8183, 0x0186, 0x018C, 0x8189, 0x0198, 0x819D, 0x8197, 0x0192,
    0x01B0, 0x81B5, 0x81BF, 0x01BA, 0x81AB, 0x01AE, 0x01A4, 0x81A1,
    0x01E0, 0x81E5, 0x81EF, 0x01EA, 0x81FB, 0x01FE, 0x01F4, 0x81F1,
    0x81D3, 0x01D6, 0x01DC, 0x81D9, 0x01C8, 0x81CD, 0x81C7, 0x01C2,
    0x0140, 0x8145, 0x814F, 0x014A, 0x815B, 0x015E, 0x0154, 0x8151,
    0x8173, 0x0176, 0x017C, 0x8179, 0x0168, 0x816D, 0x8167, 0x0162,
    0x8123, 0x0126, 0x012C, 0x8129, 0x0138, 0x813D, 0x8137, 0x0132,
    0x0110, 0x8115, 0x811F, 0x011A, 0x810B, 0x010E, 0x0104, 0x8101,
    0x8303, 0x0306, 0x030C, 0x8309, 0x0318, 0x831D, 0x8317, 0x0312,
    0x0330, 0x8335, 0x833F, 0x033A, 0x832B, 0x032E, 0x0324, 0x8321,
    0x0360, 0x8365, 0x836F, 0x036A, 0x837B, 0x037E, 0x0374, 0x8371,
    0x8353, 0x0356, 0x035C, 0x8359, 0x0348, 0x834D, 0x8347, 0x0342,
    0x03C0, 0x83C5, 0x83CF, 0x03CA, 0x83DB, 0x03DE, 0x03D4, 0x83D1,
    0x83F3, 0x03F6, 0x03FC, 0x83F9, 0x03E8, 0x83ED, 0x83E7, 0x03E2,
    0x83A3, 0x03A6, 0x03AC, 0x83A9, 0x03B8, 0x83BD, 0x83B7, 0x03B2,
    0x0390, 0x8395, 0x839F, 0x039A, 0x838B, 0x038E, 0x0384, 0x8381,
    0x0280, 0x8285, 0x828F, 0x028A, 0x829B, 0x029E, 0x0294, 0x8291,
    0x82B3, 0x02B6, 0x02BC, 0x82B9, 0x02A8, 0x82AD, 0x82A7, 0x02A2,
    0x82E3, 0x02E6, 0x02EC, 0x82E9, 0x02F8, 0x82FD, 0x82F7, 0x02F2,
    0x02D0, 0x82D5, 0x82DF, 0x02DA, 0x82CB, 0x02CE, 0x02C4, 0x82C1,
    0x8243, 0x0246, 0x024C, 0x8249, 0x0258, 0x825D, 0x8257, 0x0252,
    0x0270, 0x8275, 0x827F, 0x027A, 0x826B, 0x026E, 0x0264, 0x8261,
    0x0220, 0x8225, 0x822F, 0x022A, 0x823B, 0x023E, 0x0234, 0x8231,
    0x8213, 0x0216, 0x021C, 0x8219, 0x0208, 0x820D, 0x8207, 0x0202
  };

  for(; size; data++, size--)
    crc = (crc << 8) ^ crc_table[(crc >> 8) ^ *data];

  return crc;
}
